from datetime import datetime, timedelta
from typing import List

import requests

from ..client import client
from ..decorators import test_wrapper
from ..exceptions import FailedTest
from ..models import TestResult, ApiError
from ..utils import availability
from ..utils.adapters import get_timeslots


@test_wrapper
def test_response_format(api_url, api_key, product_id, version=1):
    '''[Timeslots] Checking response format'''
    return availability.test_response_format(api_url, api_key, product_id, 'timeslots', get_timeslots, version=1)


@test_wrapper
def test_next_30_days(api_url, api_key, product_id, version=1):
    '''[Timeslots] Checking for any availability in the next 30 days'''
    return availability.test_next_30_days(api_url, api_key, product_id, 'timeslots', get_timeslots, version=1)


@test_wrapper
def test_missing_api_key(api_url, api_key, product_id, version=1):
    '''[Timeslots] Request without API-Key'''
    return availability.test_missing_api_key(api_url, api_key, product_id, 'timeslots', version=1)


@test_wrapper
def test_incorrect_api_key(api_url, api_key, product_id, version=1):
    '''[Timeslots] Request with incorrect API-Key'''
    return availability.test_incorrect_api_key(api_url, api_key, product_id, 'timeslots', version=1)


@test_wrapper
def test_missing_argument_error(api_url, api_key, product_id, version=1):
    '''[Timeslots] Testing missing argument errors'''
    return availability.test_missing_argument_error(api_url, api_key, product_id, 'timeslots', version=1)


@test_wrapper
def test_error_for_non_existing_product(api_url, api_key, product_id, version=1):
    '''[Timeslots] Testing availability for non existing product'''
    return availability.test_error_for_non_existing_product(api_url, api_key, product_id, 'timeslots', version=1)


@test_wrapper
def incorrect_date_format(api_url, api_key, product_id, version=1):
    '''[Timeslots] Checking incorrect date format'''
    return availability.incorrect_date_format(api_url, api_key, product_id, 'timeslots', version=1)


@test_wrapper
def end_before_start_error(api_url, api_key, product_id, version=1):
    '''[Timeslots] Checking incorrect range error'''
    return availability.end_before_start_error(api_url, api_key, product_id, 'timeslots', version=1)


@test_wrapper
def past_start_date(api_url, api_key, product_id, version=1):
    '''[Timeslots] Checking past date'''
    return availability.past_start_date(api_url, api_key, product_id, 'timeslots', version=1)


@test_wrapper
def huge_date_range(api_url, api_key, product_id, version=1):
    '''[Timeslots] Checking huge date range'''
    return availability.huge_date_range(api_url, api_key, product_id, 'timeslots', version=1)


@test_wrapper
def not_allowed_method(api_url, api_key, product_id, version=1):
    '''[Timeslots] Testing methods that are not allowed'''
    return availability.not_allowed_method(api_url, api_key, product_id, 'timeslots', version=1)


@test_wrapper
def invalid_product(api_url, api_key, product_id, version=1):
    '''[Timeslots] Testing errors on invalid (non-timeslot) product'''
    return availability.test_error_for_non_timeslot_product(api_url, api_key, product_id, 'timeslots', version=1)



from supplier_api_tester.exceptions import FailedTest
from supplier_api_tester.models import TestResult


def check_api_error(api_error, expected_error) -> TestResult:
    '''Chcking if the API error matches the expected error'''
    if api_error.error_code != expected_error.error_code:
        raise FailedTest(f'Incorrect error_code ({api_error.error_code}). Expected value: {expected_error.error_code}')
    if api_error.error != expected_error.error:
        raise FailedTest(f'Incorrect error text ({api_error.error}). Expected text: {expected_error.error}')
    if not api_error.message.startswith(expected_error.message):
        return TestResult(
            status=1,
            message=f'Incorrect message text ({api_error.message}). Expected text: {expected_error.message}',
        )
    return TestResult()



import json
import requests

from typing import Callable

from .exceptions import FailedTest


def client(
    url: str,
    api_key: str,
    params: dict = None,
    headers: dict = None,
    method: Callable = requests.get,
    json_payload=None,
):
    params = params if params is not None else {}
    headers = headers if headers is not None else {'API-Key': api_key}
    try:
        response = method(url, params=params, headers=headers, json=json_payload)
    except requests.exceptions.ConnectionError:
        raise FailedTest(
            message=f'Connection error occured while testing endpoint {url}. Check if your server is available.'
        )
    except requests.exceptions.HTTPError:
        raise FailedTest(message=f'HTTP error occured while testing endpoint {url}')

    if response.status_code not in (200, 400, 403, 405, 500):
        raise FailedTest(message=f'Unexpected status code {response.status_code} from {url}')

    if response.status_code in (403, 405, 500):
        return response

    try:
        return response.json()
    except json.JSONDecodeError:
        raise FailedTest(message=f'Response from the {url} was not in a JSON format')



import click

from supplier_api_tester.tester import SupplierApiTester
from supplier_api_tester.printer import terminal_printer


def print_title(title):
    print()
    print('-' * len(title))
    print(title)
    print('-' * len(title))
    print()


@click.command()
@click.option('-u', '--url', required=True, prompt='Server URL', type=str)
@click.option('-k', '--api-key', required=True, prompt='API Key', type=str)
@click.option('-p', '--product-id', type=str, help='Product ID to call tests on. Required with -a and -t flags')
@click.option('-t', '--timeslots', is_flag=True, default=False, help='Use timeslots')
@click.option('-a', '--availability', is_flag=True, default=False, help='Run availability tests')
@click.option('-r', '--reservation', is_flag=True, default=False, help='Run reservation tests')
@click.option('-b', '--booking', is_flag=True, default=False, help='Run booking tests')
@click.option('-c', '--catalog', is_flag=True, default=False, help='Run product catalog tests')
def supplier_tester(url, api_key, product_id, timeslots, availability, reservation, booking, catalog):
    '''Test you Supplier API implementation'''

    if not any((availability, reservation, booking, catalog)):
        availability = True
        reservation = True
        booking = True
        catalog = True

    if any((availability, timeslots, reservation)) and not product_id:
        product_id = click.prompt('Product ID')

    if availability:
        print_title('AVAILABILITY TESTS')
        runner = SupplierApiTester(
            host=url,
            api_key=api_key,
            product_id=product_id,
            test_target='availability',
            timeslots=timeslots,
        )
        results = runner.run()
        terminal_printer(results)

    if reservation:
        print_title('RESERVATION TESTS')
        runner = SupplierApiTester(
            host=url,
            api_key=api_key,
            product_id=product_id,
            test_target='reservation',
            timeslots=timeslots,
        )
        results = runner.run()
        terminal_printer(results)

    if booking:
        print_title('BOOKING TESTS')
        runner = SupplierApiTester(
            host=url,
            api_key=api_key,
            product_id=product_id,
            test_target='booking',
            timeslots=timeslots,
        )
        results = runner.run()
        terminal_printer(results)

    if catalog:
        print_title('PRODUCT CATALOG')
        runner = SupplierApiTester(
            host=url,
            api_key=api_key,
            product_id=product_id,
            test_target='catalog',
            timeslots=timeslots,
        )
        results = runner.run()
        terminal_printer(results)


from datetime import datetime, timedelta, timezone
from typing import Union

import requests

from ..client import client
from ..decorators import test_wrapper
from ..exceptions import FailedTest
from ..models import TestResult, DailyAvailability, Timeslot, ApiError
from ..utils.adapters import get_reservation, get_booking, get_api_error
from ..utils.reservation import get_payload_from_slot, get_reservation_slot
from ..utils.errors import check_api_error


@test_wrapper
def test_missing_reservation_id(api_url, api_key, product_id, timeslots: bool, version=1):
    '''Booking without the reservation ID'''
    url = f'{api_url}/v{version}/booking'
    response = client(url, api_key, method=requests.post, json_payload={})
    api_error = get_api_error(response)
    expected_error = ApiError(
        error_code=1000,
        error='Missing argument',
        message="Required argument end was not found",
    )
    check_api_error(api_error, expected_error)
    return TestResult()


@test_wrapper
def test_missing_api_key(api_url, api_key, product_id, timeslots: bool, version=1):
    '''Booking without the API key'''
    url = f'{api_url}/v{version}/booking'
    response = client(url, api_key, method=requests.post, json_payload={}, headers={})

    if response.status_code != 403:
        raise FailedTest(
            f'Incorrect status code ({response.status_code}) when calling the API wihout the API-Key. '
            'Expected status code: 403.'
        )

    if response.text != 'Forbidden - Missing or incorrect API key':
        raise TestResult(
            status=1,
            message=(
                f'Incorrect text message ({response.text}). '
                'Expected message: Forbidden - Missing or incorrect API key'
            )
        )
    return TestResult()


@test_wrapper
def test_incorrect_api_key(api_url, api_key, product_id, timeslots: bool, version=1):
    '''Booking with incorrect API-Key'''
    url = f'{api_url}/v{version}/booking'
    response = client(url, api_key, method=requests.post, json_payload={}, headers={'API-Key': 'NON-EXISTING-API-KEY'})

    if response.status_code != 403:
        raise FailedTest(
            f'Incorrect status code ({response.status_code}) when calling the API wihout the API-Key. '
            'Expected status code: 403.'
        )

    if response.text != 'Forbidden - Missing or incorrect API key':
        raise TestResult(
            status=1,
            message=(
                f'Incorrect text message ({response.text}). '
                'Expected message: Forbidden - Missing or incorrect API key'
            )
        )

    return TestResult()


@test_wrapper
def test_not_allowed_method(api_url, api_key, product_id, timeslots: bool, version=1):
    '''Testing methods that are not allowed'''
    url = f'{api_url}/v{version}/booking'
    for method in (requests.get, requests.put, requests.patch, requests.delete):
        response = client(url, api_key, method=method, json_payload={})
        if response.status_code != 405:
            raise FailedTest(
                f'Incorrect status code ({response.status_code}) when calling the API via method {method}. '
                'Expected status code: 405.'
            )
    return TestResult()


@test_wrapper
def test_booking_incorrect_reservation_id(api_url, api_key, product_id, timeslots: bool, version=1):
    '''Booking with incorrect reservation ID.'''
    url = f'{api_url}/v{version}/booking'
    response = client(url, api_key, method=requests.post, json_payload={
        'reservation_id': 'NON-EXISTING-ID',
    })
    api_error = get_api_error(response)
    expected_error = ApiError(
        error_code=3002,
        error='Incorrect reservation ID',
        message="Given reservation ID is incorrect",
    )
    check_api_error(api_error, expected_error)
    return TestResult()

@test_wrapper
def test_booking(api_url, api_key, product_id, timeslots: bool, version=1):
    '''Booking tickets for at least 1 variant'''
    url = f'{api_url}/v{version}/products/{product_id}/reservation'
    slot = get_reservation_slot(api_url, api_key, product_id, timeslots)
    variant_quantity_map = {
         variant.id: 2 for variant in slot.variants if variant.max_tickets > 2
    }

    json_payload = get_payload_from_slot(slot, variant_quantity=2, min_quantity=3)
    if timeslots:
        json_payload['timeslot'] = slot.start
    response = client(url, api_key, method=requests.post, json_payload=json_payload)
    reservation = get_reservation(response)

    url = f'{api_url}/v{version}/booking'
    response = client(url, api_key, method=requests.post, json_payload={
        'reservation_id': reservation.reservation_id,
    })
    booking = get_booking(response)
    if booking.barcode_position == 'ticket':
        for variant_id, tickets_quantity in variant_quantity_map.items():
            tickets_for_variant = booking.tickets.get(variant_id)
            if tickets_for_variant is None:
                raise FailedTest(f'No tickets for variant {variant_id}')
            if len(tickets_for_variant) != tickets_quantity:
                raise FailedTest(f'Expected {tickets_quantity} codes for variant {variant_id} but got only {len(tickets_for_variant)}')

    return TestResult()



from datetime import date, datetime
from typing import List

import dacite

from ..exceptions import FailedTest
from ..models import ApiError, Booking, DailyAvailability, DailyVariants, Product, Reservation, Timeslot


def get_daily_availability(response) -> List[DailyAvailability]:
    '''Getting and testing response from the /dates endpoint'''
    if type(response) is not list:
        raise FailedTest('The response should be a JSON Array')
    try:
        days = [
            dacite.from_dict(
                data_class=DailyAvailability,
                data=day,
                config=dacite.Config(
                    type_hooks={date: date.fromisoformat},
                    strict=True,
                )
            )
            for day in response
        ]
    except (
        dacite.exceptions.WrongTypeError,
        dacite.exceptions.MissingValueError,
        dacite.exceptions.UnexpectedDataError,
    ):
        raise FailedTest('Incorrect JSON format in response from the /dates endpoint')
    return days


def get_variants(response) -> List[DailyVariants]:
    '''Getting and testing response from the /variants endpoint'''
    if type(response) is not list:
        raise FailedTest('The response should be a JSON Array')
    try:
        days = [
            dacite.from_dict(
                data_class=DailyVariants,
                data=day,
                config=dacite.Config(
                    type_hooks={date: date.fromisoformat},
                    strict=True,
                )
            )
            for day in response
        ]
    except (
        dacite.exceptions.WrongTypeError,
        dacite.exceptions.MissingValueError,
        dacite.exceptions.UnexpectedDataError,
    ):
        raise FailedTest('Incorrect JSON format in response from the /variants endpoint')
    return days


def get_timeslots(response) -> List[Timeslot]:
    '''Getting and testing response from the /timeslots endpoint'''
    if type(response) is not list:
        raise FailedTest('The response should be a JSON Array')
    try:
        days = [
            dacite.from_dict(
                data_class=Timeslot,
                data=day,
                config=dacite.Config(
                    type_hooks={date: date.fromisoformat},
                    strict=True,
                )
            )
            for day in response
        ]
    except (
        dacite.exceptions.WrongTypeError,
        dacite.exceptions.MissingValueError,
        dacite.exceptions.UnexpectedDataError,
    ):
        raise FailedTest('Incorrect JSON format in response from the /timeslots endpoint')
    return days


def get_products(response) -> List[Product]:
    '''Getting and testing response from the /products endpoint'''
    if type(response) is not list:
        raise FailedTest('The response should be a JSON Array')
    try:
        products = [
            dacite.from_dict(
                data_class=Product,
                data=product,
            )
            for product in response
        ]
    except (
        dacite.exceptions.WrongTypeError,
        dacite.exceptions.MissingValueError,
        dacite.exceptions.UnexpectedDataError,
    ):
        raise FailedTest('Incorrect JSON format in response from the /products endpoint')
    return products


def get_reservation(response) -> Reservation:
    '''Getting and testing response from the /reservation endpoint'''
    if type(response) is not dict:
        raise FailedTest('The response should be a JSON Object')
    try:
        return dacite.from_dict(
            data_class=Reservation,
            data=response,
            config=dacite.Config(
                type_hooks={datetime: datetime.fromisoformat},
                strict=True,
            )
        )
    except (
        dacite.exceptions.WrongTypeError,
        dacite.exceptions.MissingValueError,
        dacite.exceptions.UnexpectedDataError,
    ):
        raise FailedTest('Incorrect JSON format in response from the /reservation endpoint')


def get_booking(response) -> Booking:
    '''Getting and testing response from the /booking endpoint'''
    if type(response) is not dict:
        raise FailedTest('The response should be a JSON Object')
    try:
        booking = dacite.from_dict(
            data_class=Booking,
            data=response,
            config=dacite.Config(strict=True)
        )
    except (
        dacite.exceptions.WrongTypeError,
        dacite.exceptions.MissingValueError,
        dacite.exceptions.UnexpectedDataError,
    ):
        raise FailedTest('Incorrect JSON format in response from the /booking endpoint')

    if booking.barcode_format not in ('QRCODE', 'CODE128', 'CODE39', 'ITF', 'DATAMATRIX', 'EAN13'):
        raise FailedTest(f'Incorrect barcode format ({booking.barcode_format})')
    if booking.barcode_position not in ('order', 'ticket'):
        raise FailedTest(f'Incorrect value in the barcode_position field ({booking.barcode_position})')
    if booking.barcode_position == 'order' and not booking.barcode:
        raise FailedTest('Barcode for the whole order is empty')
    if booking.barcode_position == 'ticket' and not booking.tickets:
        raise FailedTest('Tickets Array is empty')
    return booking


def get_api_error(response) -> ApiError:
    '''Unpacking 400 error JSON structure'''
    if type(response) is not dict:
        raise FailedTest('400 error response should be a JSON Object')
    try:
        return dacite.from_dict(
            data_class=ApiError,
            data=response,
            config=dacite.Config(strict=True),
        )
    except (
        dacite.exceptions.WrongTypeError,
        dacite.exceptions.MissingValueError,
        dacite.exceptions.UnexpectedDataError,
    ):
        raise FailedTest(f'Incorrect response format for 400 error')



from textwrap import wrap

from terminaltables import AsciiTable


class Colors:
    OK = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'


def terminal_printer(results):
    any_message = any(bool(r.message) for r in results)
    headers = ['#', 'Time', 'Test name']
    if any_message:
        headers.append('Description')
    table_data = [headers]
    for i, result in enumerate(results, 1):
        text_color = {
            0: Colors.OK,
            1: Colors.WARNING,
            2: Colors.FAIL,
        }[result.status]
        data_row = [
            i,
            f'{text_color}{result.duration}ms{Colors.ENDC}',
            f'{text_color}{result.title}{Colors.ENDC}',
        ]
        if any_message:
            data_row.append(result.message or '')
        table_data.append(data_row)
    table = AsciiTable(table_data)
    if any_message:
        message_size = table.column_max_width(3)
        for column in table.table_data:
            column[3] = '\n'.join(wrap(column[3], message_size))
    print(table.table)



from datetime import date, datetime, timedelta

from typing import List

import requests

from supplier_api_tester.client import client
from supplier_api_tester.exceptions import FailedTest
from supplier_api_tester.models import ApiError, DailyAvailability, DailyVariants, TestResult
from supplier_api_tester.utils.adapters import get_api_error
from supplier_api_tester.utils.errors import check_api_error
from supplier_api_tester.utils.date import get_tomorrow


def test_response_format(api_url, api_key, product_id, endpoint, adapter_func, version=1):
    '''Checking response format'''
    today = datetime.utcnow().date()
    response = client(f'{api_url}/v{version}/products/{product_id}/{endpoint}', api_key, {
        'start': today.isoformat(),
        'end': today.isoformat(),
    })
    adapter_func(response)
    return TestResult()


def test_next_30_days(api_url, api_key, product_id, endpoint, adapter_func, version=1):
    '''Checking for any availability in the next 30 days'''
    start = datetime.utcnow().date()
    end = start + timedelta(days=30)

    response = client(f'{api_url}/v{version}/products/{product_id}/{endpoint}', api_key, {
        'start': start.isoformat(),
        'end': end.isoformat(),
    })
    days = adapter_func(response)

    max_tickets_sum = sum(day.max_tickets for day in days)
    if max_tickets_sum <= 0:
        raise FailedTest('There is no availability for next 30 days')

    return TestResult()


def past_start_date(api_url, api_key, product_id, endpoint, version=1):
    '''Checking availability with start date from the past'''
    today = datetime.utcnow().date()
    start = today - timedelta(days=1)
    end = today

    response = client(f'{api_url}/v{version}/products/{product_id}/{endpoint}', api_key, {
        'start': start.isoformat(),
        'end': end.isoformat(),
    })
    api_error = get_api_error(response)
    expected_error = ApiError(
        error_code=2007,
        error='Incorrect start date',
        message='Start date cannot be from the past',
    )
    return check_api_error(api_error, expected_error)


def huge_date_range(api_url, api_key, product_id, endpoint, version=1):
    '''Checking availability with huge date range'''
    today = datetime.utcnow().date()
    start = today
    end = today + timedelta(days=365 * 10)

    response = client(f'{api_url}/v{version}/products/{product_id}/{endpoint}', api_key, {
        'start': start.isoformat(),
        'end': end.isoformat(),
    })
    api_error = get_api_error(response)
    expected_error = ApiError(
        error_code=2008,
        error='Date range is too wide',
        message='Maximum date range is',
    )
    return check_api_error(api_error, expected_error)


def test_missing_api_key(api_url, api_key, product_id, endpoint, version=1):
    '''Request without API-Key'''
    tomorrow = get_tomorrow()
    response = client(f'{api_url}/v{version}/products/{product_id}/{endpoint}', api_key, {
        'start': tomorrow.isoformat(),
        'end': tomorrow.isoformat(),
    }, headers={})

    if response.status_code != 403:
        raise FailedTest(
            f'Incorrect status code ({response.status_code}) when calling the API wihout the API-Key. '
            'Expected status code: 403.'
        )

    if response.text != 'Forbidden - Missing or incorrect API key':
        raise TestResult(
            status=1,
            message=(
                f'Incorrect text message ({response.text}). '
                'Expected message: Forbidden - Missing or incorrect API key'
            )
        )

    return TestResult()


def test_incorrect_api_key(api_url, api_key, product_id, endpoint, version=1):
    '''Request with incorrect API-Key'''
    tomorrow = get_tomorrow()
    response = client(f'{api_url}/v{version}/products/{product_id}/{endpoint}', api_key, {
        'start': tomorrow.isoformat(),
        'end': tomorrow.isoformat(),
    }, headers={'API-Key': 'NON-EXISTING-API-KEY'})

    if response.status_code != 403:
        raise FailedTest(
            f'Incorrect status code ({response.status_code}) when calling the API wihout the API-Key. '
            'Expected status code: 403.'
        )

    if response.text != 'Forbidden - Missing or incorrect API key':
        raise TestResult(
            status=1,
            message=(
                f'Incorrect text message ({response.text}). '
                'Expected message: Forbidden - Missing or incorrect API key'
            )
        )

    return TestResult()


def test_missing_argument_error(api_url, api_key, product_id, endpoint, version=1):
    '''Testing missing argument errors'''

    tomorrow = get_tomorrow()
    warnings: List[str] = []

    # end is missing
    response = client(f'{api_url}/v{version}/products/{product_id}/{endpoint}', api_key, {
        'start': tomorrow.isoformat(),
    })
    api_error = get_api_error(response)
    expected_error = ApiError(
        error_code=1000,
        error='Missing argument',
        message="Required argument end was not found",
    )
    result = check_api_error(api_error, expected_error)
    if result.is_warning:
        warnings.append(result.message)

    # start is missing
    response = client(f'{api_url}/v{version}/products/{product_id}/{endpoint}', api_key, {
        'end': tomorrow.isoformat(),
    })
    api_error = get_api_error(response)
    expected_error = ApiError(
        error_code=1000,
        error='Missing argument',
        message="Required argument start was not found",
    )
    result = check_api_error(api_error, expected_error)
    if result.is_warning:
        warnings.append(result.message)

    # start and end are missing
    response = client(f'{api_url}/v{version}/products/{product_id}/{endpoint}', api_key)
    api_error = get_api_error(response)
    expected_error = ApiError(
        error_code=1000,
        error='Missing argument',
        message="Required argument start was not found",
    )
    result = check_api_error(api_error, expected_error)
    if result.is_warning:
        warnings.append(result.message)

    if warnings:
        return TestResult(
            status=1,
            message='\n '.join(warnings)
        )
    return TestResult()


def test_error_for_non_existing_product(api_url, api_key, product_id, endpoint, version=1):
    '''Testing availability for non existing product'''

    tomorrow = get_tomorrow()
    response = client(f'{api_url}/v{version}/products/NON-EXISTING-PRODUCT-ID/{endpoint}', api_key, {
        'start': tomorrow.isoformat(),
        'end': tomorrow.isoformat(),
    })
    api_error = get_api_error(response)
    expected_error = ApiError(
        error_code=1001,
        error='Missing product',
        message='Product with ID NON-EXISTING-PRODUCT-ID doesn\'t exist',
    )
    return check_api_error(api_error, expected_error)


def incorrect_date_format(api_url, api_key, product_id, endpoint, version=1):
    '''Checking incorrect date format'''
    tomorrow = get_tomorrow()
    bad_date_format = tomorrow.strftime('%d-%m-%Y')

    # start date in a bad format
    response = client(f'{api_url}/v{version}/products/{product_id}/{endpoint}', api_key, {
        'start': bad_date_format,
        'end': tomorrow.isoformat(),
    })
    api_error = get_api_error(response)
    expected_error = ApiError(
        error_code=2000,
        error='Incorrect date format',
        message=f'Incorrect date format {bad_date_format}, please use the YYYY-dd-mm format',
    )
    check_api_error(api_error, expected_error)

    # end date in a bad format
    response = client(f'{api_url}/v{version}/products/{product_id}/{endpoint}', api_key, {
        'start': tomorrow.isoformat(),
        'end': bad_date_format,
    })
    api_error = get_api_error(response)
    expected_error = ApiError(
        error_code=2000,
        error='Incorrect date format',
        message=f'Incorrect date format {bad_date_format}, please use the YYYY-dd-mm format',
    )
    check_api_error(api_error, expected_error)

    return TestResult()


def end_before_start_error(api_url, api_key, product_id, endpoint, version=1):
    '''Checking incorrect range error'''
    tomorrow = get_tomorrow()
    next_week = tomorrow + timedelta(days=7)
    response = client(f'{api_url}/v{version}/products/{product_id}/{endpoint}', api_key, {
        'start': next_week.isoformat(),
        'end': tomorrow.isoformat(),
    })
    api_error = get_api_error(response)
    expected_error = ApiError(
        error_code=2001,
        error='Incorrect date range',
        message='The end date cannot be earlier then start date',
    )
    return check_api_error(api_error, expected_error)


def not_allowed_method(api_url, api_key, product_id, endpoint, version=1):
    '''Testing methods that are not allowed'''
    tomorrow = get_tomorrow()
    for method in (requests.post, requests.put, requests.patch, requests.delete):
        response = client(f'{api_url}/v{version}/products/{product_id}/{endpoint}', api_key, {
            'start': tomorrow.isoformat(),
            'end': tomorrow.isoformat(),
        }, method=method)
        if response.status_code != 405:
            raise FailedTest(
                f'Incorrect status code ({response.status_code}) when calling the API via method {method}. '
                'Expected status code: 405.'
            )
    return TestResult()


def test_error_for_non_timeslot_product(api_url, api_key, product_id, endpoint, version=1):
    '''Testing timeslot availability for non timeslot product'''

    tomorrow = get_tomorrow()
    response = client(f'{api_url}/v{version}/products/{product_id}/{endpoint}', api_key, {
        'start': tomorrow.isoformat(),
        'end': tomorrow.isoformat(),
    })
    api_error = get_api_error(response)
    expected_error = ApiError(
        error_code=1002,
        error='Timeslot product expected',
        message=f'Requested timeslot availability for non timeslot product ID {product_id}',
    )
    return check_api_error(api_error, expected_error)


def test_error_for_timeslot_product(api_url, api_key, product_id, endpoint, version=1):
    '''Testing variant availability for timeslot product'''

    tomorrow = get_tomorrow()
    response = client(f'{api_url}/v{version}/products/{product_id}/{endpoint}', api_key, {
        'start': tomorrow.isoformat(),
        'end': tomorrow.isoformat(),
    })
    api_error = get_api_error(response)
    expected_error = ApiError(
        error_code=1003,
        error='Non-timeslot product expected',
        message=f'Requested non timeslot availability for timeslot product ID {product_id}',
    )
    return check_api_error(api_error, expected_error)


import requests

from ..client import client
from ..decorators import test_wrapper
from ..exceptions import FailedTest
from ..models import TestResult
from ..utils.adapters import get_products


@test_wrapper
def test_get_products(api_url, api_key, version=1):
    '''Get product catalog'''
    url = f'{api_url}/v{version}/products'
    response = client(url, api_key, method=requests.get)
    get_products(response)
    return TestResult()


@test_wrapper
def test_get_timeslots_products(api_url, api_key, version=1):
    '''Get product catalog with use_timeslots=true query filter'''

    url = f'{api_url}/v{version}/products'
    response = client(url, api_key, method=requests.get, params={'use_timeslots': True})
    products = get_products(response)
    _validate_timeslots(products, use_timeslots=True)
    return TestResult()


@test_wrapper
def test_get_no_timeslots_products(api_url, api_key, version=1):
    '''Get product catalog with use_timeslots=false query filter'''

    url = f'{api_url}/v{version}/products'
    response = client(url, api_key, method=requests.get, params={'use_timeslots': True})
    products = get_products(response)
    _validate_timeslots(products, use_timeslots=False)
    return TestResult()


def _validate_timeslots(products, use_timeslots):
    for product in products:
        if product.use_timeslots != use_timeslots:
            raise FailedTest(f'Product {product.id} with non matching use_timeslots returned')


from datetime import datetime, timedelta, timezone
from typing import Union

import requests

from ..client import client
from ..decorators import test_wrapper
from ..exceptions import FailedTest
from ..models import TestResult, DailyAvailability, Timeslot, ApiError
from ..utils.adapters import get_api_error
from ..utils.date import get_tomorrow
from ..utils.adapters import get_reservation
from ..utils.errors import check_api_error
from ..utils.reservation import get_payload_from_slot, get_reservation_slot


@test_wrapper
def test_missing_api_key(api_url, api_key, product_id, timeslots: bool, version=1):
    '''Request without API-Key'''
    response = client(f'{api_url}/v{version}/products/{product_id}/reservation', api_key, method=requests.post, json_payload={}, headers={})

    if response.status_code != 403:
        raise FailedTest(
            f'Incorrect status code ({response.status_code}) when calling the API wihout the API-Key. '
            'Expected status code: 403.'
        )

    if response.text != 'Forbidden - Missing or incorrect API key':
        raise TestResult(
            status=1,
            message=(
                f'Incorrect text message ({response.text}). '
                'Expected message: Forbidden - Missing or incorrect API key'
            )
        )

    return TestResult()


@test_wrapper
def test_incorrect_api_key(api_url, api_key, product_id, timeslots: bool, version=1):
    '''Request with incorrect API-Key'''

    response = client(f'{api_url}/v{version}/products/{product_id}/reservation', api_key, method=requests.post, json_payload={}, headers={
        'API-Key': 'NON-EXISTING-API-KEY',
    })

    if response.status_code != 403:
        raise FailedTest(
            f'Incorrect status code ({response.status_code}) when calling the API wihout the API-Key. '
            'Expected status code: 403.'
        )

    if response.text != 'Forbidden - Missing or incorrect API key':
        raise TestResult(
            status=1,
            message=(
                f'Incorrect text message ({response.text}). '
                'Expected message: Forbidden - Missing or incorrect API key'
            )
        )

    return TestResult()


@test_wrapper
def test_missing_argument_error(api_url, api_key, product_id, timeslots: bool, version=1):
    '''Testing missing argument errors'''
    tomorrow = get_tomorrow()
    warnings: List[str] = []
    slot = get_reservation_slot(api_url, api_key, product_id, timeslots)

    # payload is empty
    json_payload = {}
    response = client(f'{api_url}/v{version}/products/{product_id}/reservation', api_key, method=requests.post, json_payload=json_payload)
    api_error = get_api_error(response)
    expected_error = ApiError(
        error_code=1000,
        error='Missing argument',
        message="Required argument date was not found",
    )
    result = check_api_error(api_error, expected_error)
    if result.is_warning:
        warnings.append(result.message)

    # no tickets
    json_payload['date'] = tomorrow.isoformat()
    response = client(f'{api_url}/v{version}/products/{product_id}/reservation', api_key, method=requests.post, json_payload=json_payload)
    api_error = get_api_error(response)
    expected_error = ApiError(
        error_code=1000,
        error='Missing argument',
        message="Required argument tickets was not found",
    )
    result = check_api_error(api_error, expected_error)
    if result.is_warning:
        warnings.append(result.message)

    # no customer
    json_payload['tickets'] = [{
        'variant_id': slot.variants[0].id,
        'quantity': 1,
    }]
    response = client(f'{api_url}/v{version}/products/{product_id}/reservation', api_key, method=requests.post, json_payload=json_payload)
    api_error = get_api_error(response)
    expected_error = ApiError(
        error_code=1000,
        error='Missing argument',
        message="Required argument customer was not found",
    )
    result = check_api_error(api_error, expected_error)
    if result.is_warning:
        warnings.append(result.message)

    if warnings:
        return TestResult(
            status=1,
            message='\n '.join(warnings)
        )
    return TestResult()


@test_wrapper
def test_error_for_non_existing_product(api_url, api_key, product_id, timeslots: bool, version=1):
    '''Testing reservation for non-existing product'''
    url = f'{api_url}/v{version}/products/NON-EXISTING-PRODUCT-ID/reservation'
    slot = get_reservation_slot(api_url, api_key, product_id, timeslots)
    json_payload = get_payload_from_slot(slot)
    if timeslots:
        json_payload['timeslot'] = slot.start
    response = client(url, api_key, method=requests.post, json_payload=json_payload)
    api_error = get_api_error(response)
    expected_error = ApiError(
        error_code=1001,
        error='Missing product',
        message='Product with ID NON-EXISTING-PRODUCT-ID doesn\'t exist',
    )
    check_api_error(api_error, expected_error)
    return TestResult()


@test_wrapper
def test_incorrect_date_format(api_url, api_key, product_id, timeslots: bool, version=1):
    '''Testing reservation with incorrect date format'''
    url = f'{api_url}/v{version}/products/{product_id}/reservation'
    bad_date_format = '05/05/2020'
    slot = get_reservation_slot(api_url, api_key, product_id, timeslots)
    json_payload = get_payload_from_slot(slot)
    json_payload['date'] = bad_date_format
    if timeslots:
        json_payload['timeslot'] = slot.start
    response = client(url, api_key, method=requests.post, json_payload=json_payload)
    api_error = get_api_error(response)
    expected_error = ApiError(
        error_code=2000,
        error='Incorrect date format',
        message=f'Incorrect date format {bad_date_format}, please use the YYYY-dd-mm format',
    )
    check_api_error(api_error, expected_error)
    return TestResult()


@test_wrapper
def test_past_date(api_url, api_key, product_id, timeslots: bool, version=1):
    '''Testing reservation with past date'''
    url = f'{api_url}/v{version}/products/{product_id}/reservation'
    slot = get_reservation_slot(api_url, api_key, product_id, timeslots)
    json_payload = get_payload_from_slot(slot)
    yesterday = datetime.utcnow().date() - timedelta(days=1)
    json_payload['date'] = yesterday.isoformat()
    if timeslots:
        json_payload['timeslot'] = slot.start
    response = client(url, api_key, method=requests.post, json_payload=json_payload)
    api_error = get_api_error(response)
    expected_error = ApiError(
        error_code=2009,
        error='Incorrect date',
        message=f'Cannot use the past date',
    )
    check_api_error(api_error, expected_error)
    return TestResult()


@test_wrapper
def test_not_allowed_method(api_url, api_key, product_id, timeslots: bool, version=1):
    '''Testing methods that are not allowed'''
    url = f'{api_url}/v{version}/products/{product_id}/reservation'
    slot = get_reservation_slot(api_url, api_key, product_id, timeslots)
    json_payload = get_payload_from_slot(slot)
    if timeslots:
        json_payload['timeslot'] = slot.start
    for method in (requests.get, requests.put, requests.patch, requests.delete):
        response = client(url, api_key, method=method, json_payload=json_payload)
        if response.status_code != 405:
            raise FailedTest(
                f'Incorrect status code ({response.status_code}) when calling the API via method {method}. '
                'Expected status code: 405.'
            )
    return TestResult()


@test_wrapper
def test_reservation(api_url, api_key, product_id, timeslots: bool, version=1):
    '''Reserving tickets for at least 1 variant'''
    url = f'{api_url}/v{version}/products/{product_id}/reservation'
    slot = get_reservation_slot(api_url, api_key, product_id, timeslots)
    json_payload = get_payload_from_slot(slot)
    if timeslots:
        json_payload['timeslot'] = slot.start
    response = client(url, api_key, method=requests.post, json_payload=json_payload)
    reservation = get_reservation(response)
    if not reservation.reservation_id:
        raise FailedTest('No reservation ID found')
    if reservation.expires_at.tzname() != 'UTC':
        raise FailedTest('Expiration time should be in the UTC timezone.')
    if datetime.now(timezone.utc) + timedelta(minutes=14) > reservation.expires_at:
        raise FailedTest('Reservation should be held at least 15 minutes.')
    return TestResult()
